using System;
using System.Drawing;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Data;
using System.Runtime.InteropServices;
using System.Threading;
using System.Text;

namespace GP32DeMon
{
	/// <summary>
	/// GP32 USB debug console
	/// </summary>
	public class fDeMon : System.Windows.Forms.Form
	{
		public IntPtr mGP32USBHandle;
		private Thread mInputPollThread;
		volatile private Boolean mPollInput;

		// Torlus - output buffer
		public byte [] outputBuffer = new byte[256];
		public int outputLen = 0;

		GP32USB.Overlapped ovr;

		private System.Windows.Forms.TextBox hwOutput;
		/// <summary>
		/// Required designer variable.
		/// </summary>
		private System.ComponentModel.Container components = null;

		/// <summary>
		/// constructor, kicks off the polling thread
		/// </summary>
		public fDeMon()
		{
			InitializeComponent();
		
			// kick off input watcher
			mPollInput = true;
			mInputPollThread = new Thread (new ThreadStart (this.InputPollThreadFn) );
			mInputPollThread.Start();
		}

		/// <summary>
		/// thread function sitting in the background polling the GP32 for input
		/// </summary>
		public void InputPollThreadFn()
		{
			Byte [] inBuf = new Byte[512];
			
			ovr.EventHandle = GP32USB.CreateEvent(IntPtr.Zero, true, false, "IOEvent");
			ovr.InternalHigh = ovr.InternalLow = 0;
			ovr.OffsetHigh = ovr.OffsetLow = 0;
			Int32 wait;
			UInt32 bytes;
          
			Thread.Sleep(250);

			while (mPollInput)
			{
				// try to read from the device
				GP32USB.ResetEvent(ovr.EventHandle);
				GP32USB.GP32USBRead(
					mGP32USBHandle,
					inBuf,
					256,
					ref ovr);

				// sync
				wait = GP32USB.WaitForSingleObject(
					ovr.EventHandle, 
					500);

				if (wait == 0)
				{
					// pause to get the result of the op
					bytes = 0;
					GP32USB.GetOverlappedResult(
						GP32USB.GP32USBGetDeviceHandle(mGP32USBHandle), 
						ref ovr, 
						out bytes, 
						true);

					// have bytes arrived?
					if (bytes != 0)
					{
						// append the (assumed) text to the debug console
						
						//Torlus - packet prefixed by a 1-byte length
						bytes = inBuf[0];
						String text = Encoding.ASCII.GetString(inBuf);
						text = text.Substring(1,(int)bytes);
						hwOutput.AppendText(text);

						// send the sync back to the GP32 so it can carry on working

						//Torlus - send outputBuffer as sync
						outputBuffer[0] = (byte)outputLen;
						
						GP32USB.ResetEvent(ovr.EventHandle);
						GP32USB.GP32USBWrite(
							mGP32USBHandle,
							outputBuffer,
							256,
							ref ovr);

						outputLen = 0;

						wait = GP32USB.WaitForSingleObject(
							ovr.EventHandle, 
							1000);

						if (wait == 0)
						{
							bytes = 0;
							GP32USB.GetOverlappedResult(
								GP32USB.GP32USBGetDeviceHandle(mGP32USBHandle), 
								ref ovr, 
								out bytes, 
								true);
						}

					} /* bytes != 0 */
				} /* wait == 0 */
			}
		}

		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			if( disposing )
			{
				if (components != null) 
				{
					components.Dispose();
				}
			}
			base.Dispose( disposing );
		}

		#region Windows Form Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		private void InitializeComponent()
		{
			System.Resources.ResourceManager resources = new System.Resources.ResourceManager(typeof(fDeMon));
			this.hwOutput = new System.Windows.Forms.TextBox();
			this.SuspendLayout();
			// 
			// hwOutput
			// 
			this.hwOutput.BackColor = System.Drawing.Color.FromArgb(((System.Byte)(0)), ((System.Byte)(64)), ((System.Byte)(0)));
			this.hwOutput.Dock = System.Windows.Forms.DockStyle.Fill;
			this.hwOutput.Font = new System.Drawing.Font("Courier New", 9F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((System.Byte)(0)));
			this.hwOutput.ForeColor = System.Drawing.Color.FromArgb(((System.Byte)(192)), ((System.Byte)(255)), ((System.Byte)(192)));
			this.hwOutput.Location = new System.Drawing.Point(0, 0);
			this.hwOutput.MaxLength = 65535;
			this.hwOutput.Multiline = true;
			this.hwOutput.Name = "hwOutput";
			this.hwOutput.ReadOnly = true;
			this.hwOutput.ScrollBars = System.Windows.Forms.ScrollBars.Both;
			this.hwOutput.Size = new System.Drawing.Size(656, 422);
			this.hwOutput.TabIndex = 0;
			this.hwOutput.Text = "";
			this.hwOutput.KeyPress += new System.Windows.Forms.KeyPressEventHandler(this.hwOutput_KeyPress);
			this.hwOutput.TextChanged += new System.EventHandler(this.hwOutput_TextChanged);
			// 
			// fDeMon
			// 
			this.AutoScaleBaseSize = new System.Drawing.Size(5, 14);
			this.ClientSize = new System.Drawing.Size(656, 422);
			this.Controls.Add(this.hwOutput);
			this.Font = new System.Drawing.Font("Tahoma", 8.25F, System.Drawing.FontStyle.Regular, System.Drawing.GraphicsUnit.Point, ((System.Byte)(0)));
			this.Icon = ((System.Drawing.Icon)(resources.GetObject("$this.Icon")));
			this.Name = "fDeMon";
			this.Text = "PandaForth USB Console - based on GP32 Debug Monitor -- www.ishani.plus.com/gp32";
			this.Closed += new System.EventHandler(this.fDeMon_Closing);
			this.ResumeLayout(false);

		}
		#endregion

		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[MTAThread]
		static void Main() 
		{
			// create new instance of monitor, try to link to GP32
			fDeMon GP32DeMon = new fDeMon();
			GP32DeMon.mGP32USBHandle = GP32USB.GP32USBConnect(true);

			// got a USB connection to GP32?
			if (GP32DeMon.mGP32USBHandle == new IntPtr(-1))
			{
				MessageBox.Show("Could not connect to the GP32!\n\nError Report :\n" +
					GP32USB.FormatGP32USBError(GP32USB.GP32USBGetLastError()),
					"GP32 Debug Monitor",
					MessageBoxButtons.OK,
					MessageBoxIcon.Error);

				return;
			}

			// run the monitor
			Application.Run(GP32DeMon);
			Environment.Exit(0);
		}

		private void fDeMon_Closing(object sender, System.EventArgs e)
		{
			// kill input polling thread
			mInputPollThread.Abort();

			// unplug from USB
			GP32USB.GP32USBDisconnect(mGP32USBHandle);
		}

		private void hwOutput_KeyPress(object sender, System.Windows.Forms.KeyPressEventArgs e)
		{
			// Torlus - output buffer
			outputBuffer[1+outputLen++] = (byte)e.KeyChar;
		}

		private void hwOutput_TextChanged(object sender, System.EventArgs e)
		{
		
		}
	}


	/// <summary>
	/// GP32USB DLL import class
	/// provides the platform invoke commands to call
	/// the functions from the DLL
	/// </summary>
	public class GP32USB
	{
		[ StructLayout( LayoutKind.Sequential )]
		public struct Overlapped
		{
			public Int32 InternalLow;
			public Int32 InternalHigh;
			public Int32 OffsetLow;
			public Int32 OffsetHigh;
			public IntPtr EventHandle;
		}

		public enum GP32USBError
		{
			GP32_NOERROR,
			GP32_CANNOT_FIND_DEVICE,
			GP32_CANNOT_GET_INFO,
			GP32_CANNOT_GET_DETAILS,
			GP32_CANNOT_CREATE_CHANNEL
		}


		[DllImport("kernel32.dll", SetLastError=true)]
		static public extern Boolean GetOverlappedResult(
			IntPtr hFile, 
			ref Overlapped ovr,
			out UInt32 nNumberOfBytesTransferred, 
			Boolean bWait);

		[DllImport("kernel32.dll")]
		static public extern Int32 WaitForSingleObject(
			IntPtr handle, 
			int timeOut);

		[DllImport("kernel32.dll")]
		static public extern Boolean ResetEvent(
			IntPtr handle);

		[DllImport("kernel32.dll")]
		static public extern IntPtr CreateEvent(
			IntPtr eventAttributes, 
			bool manualReset, 
			bool initialState, 
			String name);


		[DllImport("GP32USB.dll")]
		static public extern GP32USBError GP32USBGetLastError();
		
		[DllImport("GP32USB.dll")]
		static public extern IntPtr GP32USBConnect(
			Boolean overlapped);

		[DllImport("GP32USB.dll")]
		static public extern IntPtr GP32USBGetDeviceHandle(
			IntPtr gpusb);

		[DllImport("GP32USB.dll")]
		static public extern void GP32USBDisconnect(
			IntPtr gpusb);

		[DllImport("GP32USB.dll")]
		static public extern Int32 GP32USBRead(
			IntPtr gpusb,
			byte[] buf,
			Int32 bufLen,
			ref Overlapped ovr);
	
		[DllImport("GP32USB.dll")]
		static public extern Int32 GP32USBWrite(
			IntPtr gpusb,
			byte[] buf,
			Int32 bufLen,
			ref Overlapped ovr);

		/// <summary>
		/// little helper function to generate a readible error from
		/// an error code generated by GP32USB
		/// </summary>
		static public String FormatGP32USBError(
			GP32USBError err)
		{
			switch (err)
			{
				case GP32USBError.GP32_NOERROR:
					return "No Error";

				case GP32USBError.GP32_CANNOT_FIND_DEVICE:
					return "Cannot find any suitable USB devices : have you installed the GP32 USB driver?";

				case GP32USBError.GP32_CANNOT_GET_INFO:
				case GP32USBError.GP32_CANNOT_GET_DETAILS:
					return "Cannot examine USB devices : GP32 may be turned off, or USB cable not connected properly";

				case GP32USBError.GP32_CANNOT_CREATE_CHANNEL:
					return "Could not create channel to talk to USB device";
			}
			return "Unknown Error";
		}
	}

}
